#!/bin/bash
# example revssh client

# stop on nonzero exit
set -e


# show some help
function usage
{
    echo "Usage: $0 <satellite base domain> <dest host> <dest port> <input directory>" 1>&2
    exit
}

# see if we need to show some help
if [[ "${#@}" -ne "4" ]]; then
  usage
fi

# since we're laying down an ssh key, be very restrictive so
# it's not accidentally world-readable
umask 0077

BASEDOM="${1}"
HOST="${2}"
PORT="${3}"
INPUT_DIR="${4}"

LOCK_DIR="${INPUT_DIR}/.locks"

if [[ ! -d "${INPUT_DIR}" ]]; then
  echo "Input directory does not exist." 1>&2
  exit 1
fi

mkdir -p "${LOCK_DIR}"

# Process a new token file
function process_new_file
{
  local FILE="${1}"
  local IDDIR=$( mktemp -d /dev/shm/revssh-client.XXXXXX )

  # Parse token from the token file name: <token>.token
  local TOKEN=$( basename "${FILE}" | cut -d. -f1 )

  # Redeem the token and start a reverse SSH tunnel
  echo "Redeeming token: ${TOKEN} | IDDIR: $IDDIR"
  echo "https://${BASEDOM}/redeemtoken_v2.cgi?jobid=${TOKEN}&revssh=1"
  curl --silent "https://${BASEDOM}/redeemtoken_v2.cgi?jobid=${TOKEN}&revssh=1" > "${IDDIR}/redeem"
  echo "finished request"
  if [[ $( grep -c '^Success' "${IDDIR}/redeem" ) -ne 1 ]]; then
    echo "Failed to redeem token. Message below..." 1>&2
    cat "${IDDIR}/redeem" 1>&2
    cd ..
    cleanup
    exit 1
  fi
  echo "Redeemed token, proceeding to establish ssh tunnel"

  # get the username and key
  head -2 "${IDDIR}/redeem" | tail -1 > "${IDDIR}/username"
  tail -n +3 "${IDDIR}/redeem" > "${IDDIR}/privkey"

  # establish the ssh tunnel
  ssh -p 23 -i "${IDDIR}/privkey" "$(cat "${IDDIR}/username")@${BASEDOM}" -o TCPKeepAlive=yes -o ServerAliveInterval=60 -o StrictHostKeyChecking=accept-new -N -R "/socket/tunnel:${HOST}:${PORT}" &

  # Store the PID and IDDIR in a lock file
  echo "${!},${IDDIR}" > "${LOCK_DIR}/$(basename "${FILE}").lock"
}

# Remove a reverse SSH tunnel connection
function remove_connection
{
  local LOCK_FILE="${1}"
  local PID=$(cut -d, -f1 "${LOCK_FILE}")
  local IDDIR=$(cut -d, -f2 "${LOCK_FILE}")

  echo "Removing connection for lock file ${LOCK_FILE}..."

  # Kill the SSH tunnel process and delete the IDDIR
  kill "$PID"
  rm -rf "${IDDIR}"

  # Remove the lock file
  rm "${LOCK_FILE}"
}

# cleanup code
cleanup() {
  # on exit, remove all reverse ssh tunnel connections
  echo "Cleaning up all reverse ssh tunnel connections..."
  for LOCK_FILE in "${LOCK_DIR}"/*.lock; do
    if [[ -f "${LOCK_FILE}" ]]; then
      TOKEN_FILE="${INPUT_DIR}/$(basename "${LOCK_FILE}" .lock)"
      rm $TOKEN_FILE # delete the token file
      remove_connection "${LOCK_FILE}" # remove its connection
    fi
  done
  echo "Done!"
  exit
}
trap cleanup INT TERM EXIT


# check if inotifywait is available
if command -v inotifywait >/dev/null 2>&1; then
  echo "Using inotifywait to watch for file changes."

  # Process new token files
  for FILE in "${INPUT_DIR}"/*.token; do
    if [[ -f "${FILE}" && ! -f "${LOCK_DIR}/$(basename "${FILE}").lock" ]]; then
      process_new_file "${FILE}"
    fi
  done

  # Remove deleted token files' connections
  for LOCK_FILE in "${LOCK_DIR}"/*.lock; do
    if [[ -f "${LOCK_FILE}" ]]; then
      TOKEN_FILE="${INPUT_DIR}/$(basename "${LOCK_FILE}" .lock)"
      if [[ ! -f "${TOKEN_FILE}" ]]; then
        remove_connection "${LOCK_FILE}"
      fi
    fi
  done


  # Use inotify to watch for file changes
  while inotifywait -e create,delete "${INPUT_DIR}"; do
    # Process new token files
    for FILE in "${INPUT_DIR}"/*.token; do
      if [[ -f "${FILE}" && ! -f "${LOCK_DIR}/$(basename "${FILE}").lock" ]]; then
        process_new_file "${FILE}"
      fi
    done

    # Remove deleted token files' connections
    for LOCK_FILE in "${LOCK_DIR}"/*.lock; do
      if [[ -f "${LOCK_FILE}" ]]; then
        TOKEN_FILE="${INPUT_DIR}/$(basename "${LOCK_FILE}" .lock)"
        if [[ ! -f "${TOKEN_FILE}" ]]; then
          remove_connection "${LOCK_FILE}"
        fi
      fi
    done
  done
else
  # Fallback to the original loop approach
  echo "inotifywait not found. Falling back to loop approach."
  while true; do
    # Process new token files
    for FILE in "${INPUT_DIR}"/*.token; do
      if [[ -f "${FILE}" && ! -f "${LOCK_DIR}/$(basename "${FILE}").lock" ]]; then
        process_new_file "${FILE}"
      fi
    done

    # Remove deleted token files' connections
    for LOCK_FILE in "${LOCK_DIR}"/*.lock; do
      if [[ -f "${LOCK_FILE}" ]]; then
        TOKEN_FILE="${INPUT_DIR}/$(basename "${LOCK_FILE}" .lock)"
        if [[ ! -f "${TOKEN_FILE}" ]]; then
          remove_connection "${LOCK_FILE}"
        fi
      fi
    done

    # Sleep for 5 seconds
    sleep 5
  done
fi
