#!/bin/bash
mkdir -p outputs/
touch "$(pwd)/filebrowser.db"

is_executable_app=$1
setup_app_proxying=$2

# Input validation
# trust is nice but let's make sure something didn't go wrong
# satellite_host: hostname x.y.z
satellite_host=$3
re="^[a-z0-9\-\.]+$"
if ! [[ $satellite_host =~ $re ]]; then
  echo The satellite hostname has invalid characters
  exit 2
fi
# 1-63 characters between each dot
IFS="."
for label in ${satellite_host[@]}; do
  if [ ${#label} -gt 63 ]; then
    echo The satellite hostname has a label that is too long
    exit 3
  fi
  if [ ${#label} -lt 1 ]; then
    echo The satellite hostname has consecutive dots
    exit 4
  fi
done
# reset IFS to default
IFS=$' \t\n'

# job_proxy_id: a-zA-Z0-9 and dashes
re="^[a-zA-Z0-9\-]+$"
job_proxy_id=$4
if ! [[ $job_proxy_id =~ $re ]]; then
  echo The satellite job_proxy_id has invalid characters
  exit 5
fi

# shift to remove the first four arguments
shift
shift
shift
shift

# Function to parse env vars from a .env file, line by line.
# This returns a string of quoted env var keys/values (var1="a" var2="b", etc.)
# Note: We quote all env values, since Tapis doesn't quote any values
parse_env_file() {
    local env_file="$1"
    local env_string=""

    while IFS='=' read -r key value; do
        if [[ ! $key =~ ^[[:space:]]*# && ! -z $key && ! -z $value ]]; then
            # note: we're quoting the values here, since tapis doesn't quote any values
            env_string+="$key=\"$value\" "
        fi
    done < "$env_file"

    echo "$env_string"
}

# Function to build a command string with environment variables after parsing them from the input
extract_executable_cmd() {
  # Store the input command as a string
  local input_cmd="$@"

  # Check if the input_cmd starts with docker or singularity
  if [[ $input_cmd == "docker"* ]]; then
    # Extract the program name and its arguments from the input command
    local prog_and_args=$(echo "$input_cmd" | grep -oP '(?<=osp_placeholder ).*')

    # Return the combined command string
    echo "$prog_and_args"
  elif [[ $input_cmd == "singularity"* ]]; then
    # Extract the program name and its arguments from the input command
    local prog_and_args=$(echo "$input_cmd" | grep -oP '(?<=osp_placeholder.sif ).*')

    # Return the combined command string
    echo "$prog_and_args"
  fi
}
extract_executable_env_vars() {
  # Store the input command as a string
  local input_cmd="$@"

  # Check if the input_cmd starts with docker or singularity
  if [[ $input_cmd == "docker"* ]]; then
    # If it starts with docker, replace the placeholder with the docker image name
    # Parse env vars from tapisjob.env file (we only consider case where tapisjob.env is present, since tapis doesn't run docker on batch systems, where the env vars would be passed in as args, instead of in a .env file)
    local env_vars=$(parse_env_file "$(pwd)/tapisjob.env")

    # Return the combined command string
    echo "$env_vars"
  elif [[ $input_cmd == "singularity"* ]]; then
    # If it starts with singularity, replace the placeholder with the singularity image name
    local env_vars=$(echo "$input_cmd" | grep -oP '(?<=--env ).*?(?= osp_placeholder.sif)' | tr ',' ' ')

    # Return the combined command string
    echo "$env_vars"
  fi
}

# echo "Executable app env vars: $executable_app_envvars"
# echo "Executable app command: $executable_app_cmd"

# get all the args as a single string (except the first two args)
container_cmd="$@"

# env_vars=$(parse_env_file "$(pwd)/tapisjob.env")
# executable_app_envvars=$(extract_executable_env_vars "$container_cmd")
# executable_app_cmd=$(extract_executable_cmd "$container_cmd")
# echo "env_vars: $env_vars" >> ./debug.txt
# echo "executable_app_envvars: $executable_app_envvars" >> ./debug.txt
# echo "executable_app_cmd: $executable_app_cmd" >> ./debug.txt

# Check if a tapisjob.env file exists
if [ -r "$(pwd)/tapisjob.env" ] ; then
  # Read the JOB_ID from the tapisjob.env file
  JOB_ID=$(grep -oP '^_tapisJobUUID=\K.*$' "$(pwd)/tapisjob.env")
else
  # Read the container command arguments (note: tapisjob.env is not always present for all fork + batch jobs)
  if [[ $container_cmd == "docker"* ]]; then
    # I don't think we'll encounter this case with Tapis, since this can only happen with Docker + BATCH jobs, but i don't think Tapis currently supports this config.
    name_arg=$(echo "$container_cmd" | grep -o -E -- '--name [^ ]+' | head -n 1)
    JOB_ID=${name_arg#--name }  # remove "--name " prefix
  elif [[ $container_cmd == "singularity"* ]]; then
    # extract the value of the _tapisJobUUID env var from the singularity cmd
    # it will be _tapisJobUUID=<job_id> somewhere in the cmd
    # this has NOTHING to do with osp_placeholder
    JOB_ID=$(echo "$container_cmd" | grep -oP '_tapisJobUUID=[\w-]*' | cut -d'=' -f2)
  fi
fi

PID_FILE="$(pwd)/${JOB_ID}.pid"
echo "pid file: $PID_FILE" >> ./debug.txt

# edit the singularity cmd to add in the --pid-file option
# this is invalid (pid file only available for instance start): singularity_cmd=${singularity_cmd/singularity run/singularity run --pid-file ${PID_FILE}}
container_cmd=${container_cmd/singularity instance start/singularity instance start --pid-file "${PID_FILE}"}
container_cmd=${container_cmd/docker run/docker run --network ${JOB_ID}}
# echo "Container command: $container_cmd"

if [ "$is_executable_app" == "1" ]
then
  executable_app_envvars=$(extract_executable_env_vars "$container_cmd")
  executable_app_cmd=$(extract_executable_cmd "$container_cmd")
else
  executable_app_cmd=""
fi

# TODO: Auto compute the path for this, instead of manually specifying it explicitly
CURRENT_SCRIPT_DIR=$(dirname "$0")
CLEANUP_SCRIPT_PATH="$CURRENT_SCRIPT_DIR/osp_launcher.clean.sh"
EXEC_CLEANUP_SCRIPT_PATH="$CURRENT_SCRIPT_DIR/osp_launcher.exec_clean.sh"

find_unused_local_port() {
  local i port
  local -A p

  netstat -aln | awk '
    $6 == "LISTEN" {
      if ($4 ~ "[.:][0-9]+$") {
        split($4, a, /[:.]/)
        port = a[length(a)]
        p[port] = 1
      }
    }
    END {
      for (i = 3000; i < 65000 && p[i]; i++){};
      if (i == 65000) {exit 1};
      print i
    }
  '
}

# Define the OSP_APP_PORT
if [ "$setup_app_proxying" == "1" ]
then
	if [[ $container_cmd == "docker"* ]]; then
		# if we're running a docker container, parse the OSP_APP_PORT from the env vars file tapis generated
		OSP_APP_PORT=$(grep -oP '^OSP_APP_PORT=\K.*$'  "$(pwd)/tapisjob.env")
    		container_cmd=${container_cmd/docker run/docker run --expose $OSP_APP_PORT}
	else
		# otherwise, find an open port
		OSP_APP_PORT=$(find_unused_local_port)
	fi
else
	# OSP_APP_PORT is set to -1 for cases where it's not used (non-web/vnc app, etc.)
	OSP_APP_PORT="-1"
fi


echo "Is executable app: ${is_executable_app}, Setup app proxying: ${setup_app_proxying}, Job proxy id: ${job_proxy_id}, Job ID: ${JOB_ID}, Satellite host: ${satellite_host}, OSP app port: $OSP_APP_PORT" > ./ospjob_proxy_info.txt

# If we need to setup app proxying,
# run revssh-client in the background
if [[ $container_cmd == "docker"* ]]; then
  if [ "$setup_app_proxying" == "1" ]
  then
    # create the docker network
    docker network create ${JOB_ID} > /dev/null 2>&1
  fi
fi

if [ "$setup_app_proxying" == "1" ]
then
	# echo "running revssh-client in the bg"
	mkdir -p ./.revssh
  	mkdir -p ./.revssh/.locks
	
	# auto-create the token file for this initial job proxy id
	touch ./.revssh/${job_proxy_id}.token
	
	# start the reverse ssh tunnel client in the bg
  	docker run -d --rm --network ${JOB_ID} \
                  --name=revssh-${JOB_ID} \
                  -v "$(pwd)/.revssh":/.revssh \
                  -it revssh-client:v5 \
                  $satellite_host $JOB_ID $OSP_APP_PORT /.revssh > ./revssh-container-rund.txt 2>&1
fi


# Function to replace the value for an option with a quoted value
quote_value() {
    local input_string="$1"
    local option="$2"

    echo "$input_string" | awk -v opt="$option" '
    {
        for (i=1; i<=NF; i++) {
            if ($i == opt) {
                # Join the next fields until we hit an option or end of line
                val=""
                for (j=i+1; j<=NF; j++) {
                    if ($j ~ /^--/) break
                    if (val != "") val = val " " $j
                    else val = $j
                }
                printf "%s \"%s\" ", opt, val
                i = j-1  # Skip ahead
            } else {
                printf "%s ", $i
            }
        }
        print ""
    }'
}

# is this an executable app?
if [ "$is_executable_app" == "1" ]
then
    # echo "executable app strategy"

    # Check if the container cmd runs in the background or the foreground.
    # 1. if it's a docker cmd, check if it has the -d flag. If it does, then it's a background cmd and vice versa.
    # 2. if it's a singularity cmd, check if it has "instance start" in it, instead of "run". If it has "start", then it's a background cmd and vice versa.

    is_bg_cmd=0
    if [[ $container_cmd == "docker"* ]]; then
        # check if the docker cmd has the -d flag
        if [[ $container_cmd == *" -d "* ]]; then
            is_bg_cmd=1
        fi
    elif [[ $container_cmd == "singularity instance start"* ]]; then
        is_bg_cmd=1
    fi

    # For background cmds:
    #     1. Run the executable app command in the bg with nohup.
    #     2. Run a container in the background with the name set to the job id and this container should start/stop with the executable command bg process.
    #   For example, whenever the executable app is killed, the container should also be killed.
    #   And whenever the bg container is killed, the executable app should also be killed.
    # For foreground cmds:
    #     1. Run the executable app command in the fg

    if [[ "$is_bg_cmd" -eq 1 ]]
    then
      # run the executable cmd in the bg
      # echo "running job container in bg..."
      nohup env "OSP_APP_PORT=$OSP_APP_PORT $executable_app_envvars" $executable_app_cmd > outputs/ospjob.log 2>&1 &
      exec_job_pid="$!"

      # run a container in the bg with the name set to the job id
      IS_DOCKER="0"
      if [[ $container_cmd == "docker"* ]]
      then
        IS_DOCKER="1"
        CID_FILE="$(pwd)/${JOB_ID}.cid"

        # busybox doesn't have trap command
        # docker run -d --name testb --user $(id -u):$(id -g) --init -it debian bash -c 'trap "exit 0" SIGTERM SIGINT; while true; do sleep 1; done'
        # docker run -d --name $JOB_ID --user $(id -u):$(id -g) --cidfile $CID_FILE -it busybox
        docker run -d --name $JOB_ID --user $(id -u):$(id -g) --cidfile "$CID_FILE" --init -it debian bash -c 'trap "exit 0" SIGTERM SIGINT; while true; do sleep 1; done'
      elif [[ $container_cmd == "singularity instance start"* ]]; then
        singularity instance start docker://busybox $JOB_ID
      fi

      # If we setup app proxying, run the cleanup script
      if [ "$setup_app_proxying" == "1" ]
      then
        nohup $CLEANUP_SCRIPT_PATH $IS_DOCKER $JOB_ID > ${JOB_ID}_cleanup.txt 2>&1 &
      fi

      # Start the exec cleanup scripts in the bg
      # these processes make sure that the job container is killed when the executable app process is killed,
      # and vice versa
      nohup $EXEC_CLEANUP_SCRIPT_PATH 0 $exec_job_pid $JOB_ID $IS_DOCKER > ${JOB_ID}_exec_cleanup1.txt 2>&1 &
      nohup $EXEC_CLEANUP_SCRIPT_PATH 1 $exec_job_pid $JOB_ID $IS_DOCKER > ${JOB_ID}_exec_cleanup2.txt 2>&1 &
    else
      # echo "running job container in fg..."
      OSP_APP_PORT=$OSP_APP_PORT $executable_app_envvars $executable_app_cmd > job.log 2>&1

      # If we setup app proxying, run the cleanup script
      if [ "$setup_app_proxying" == "1" ]
      then
        # cleanup
        # echo "Cleaning up..."
        
        # This will remove the revssh-client process
        kill $REVSSH_CLIENT_PID
      fi
    fi

elif [[ $container_cmd == "docker"* ]]; then
	# This is the only case where this launcher script will not block until the job container is done
	# We only need to block when using SLURM, but Tapis doesn't support Docker + Slurm (only Docker on FORK systems)
	# and on FORK systems, it's ok if we exit after running the job docker container with -d

  	echo "Container command: $container_cmd" >> ./debug.txt
	# Update the string for --cidfile and --env-file options
	container_cmd=$(quote_value "$container_cmd" "--cidfile")
	container_cmd=$(quote_value "$container_cmd" "--env-file")
	echo "Updated container command: $container_cmd" >> ./debug.txt
  	
	# run the docker command
	CID_FILE="$(pwd)/${JOB_ID}.cid"
	echo "cid file: $CID_FILE" >> ./debug.txt
	# ${container_cmd} > /dev/null 2>&1
	eval "${container_cmd}" > /dev/null 2>&1
	
	# create the cid file
	if [ -r "${CID_FILE}" ] ; then
		CONTAINER_ID=$(cat "${CID_FILE}")
	else
		# Write container id to cid file
		CONTAINER_ID=$(docker container ls --all --quiet --no-trunc --filter "name=^${JOB_ID}$")
		echo ${CONTAINER_ID} > "${CID_FILE}"
	fi

	echo "Container ID: $CONTAINER_ID" >> ./debug.txt

	# capture Docker process stdout/stderr log
	nohup docker logs --follow ${CONTAINER_ID} > outputs/ospjob.log 2>&1 &

	# start the cleanup process in the background
  	nohup $CLEANUP_SCRIPT_PATH 1 ${JOB_ID} > ${JOB_ID}_cleanup.txt 2>&1 &

	# echoing the container id here, since that's what tapis's default launcher script ends up doing
	echo $CONTAINER_ID

elif [[ $container_cmd == "singularity instance start"* ]]; then
    # echo "singularity instance start strategy"

    # If we need to setup app proxying,
    # launch the clean up process in the bg
    if [ "$setup_app_proxying" == "1" ]
    then
	    # echo "Running clean up script in background..."
	    $CLEANUP_SCRIPT_PATH 0 ${JOB_ID} "${PID_FILE}" $REVSSH_CLIENT_PID > ${JOB_ID}_cleanup.txt 2>&1 &
	    CLEANUP_SCRIPT_PID=$!
    fi

    # run this container cmd
    # echo "running job container..."
    OSP_APP_PORT=$OSP_APP_PORT ${container_cmd} > outputs/ospjob.log 2>&1

    # If we setup app proxying, run the cleanup script
    if [ "$setup_app_proxying" == "1" ]
    then
	    # wait for the cleanup process to exit
	    # We need to do this, cuz otherwise this script will just end right after starting the singularity container in the bg, and the slurm job will be cancelled!
	    # echo "waiting for the cleanup script to exit..."
	    while kill -0 $CLEANUP_SCRIPT_PID > /dev/null 2>&1; do
		    sleep 1
	    done
    fi

elif [[ $container_cmd == "singularity run"* ]]; then
    # echo "singularity run strategy"
    
    # run this container cmd
    # echo "running job container..."
    OSP_APP_PORT=$OSP_APP_PORT ${container_cmd} > outputs/ospjob.log 2>&1

    # If we setup app proxying, run the cleanup script
    if [ "$setup_app_proxying" == "1" ]
    then
	    # cleanup
	    # echo "Cleaning up..."
	    
	    # This will remove the revssh-client process
	    kill $REVSSH_CLIENT_PID
    fi
# else
    # echo "Invalid singularity command!"
fi
